import os.path

import flatbuffers
import sh
from typing import Any, Dict, Optional

from .server import Server
from .settings import Partition
from .interface.module_interface import ModuleInterface
from .interface.method_request import MethodRequest
from .interface.schema import get_handlers, get_handler
from .interface.base import MethodRequestArg
from .response import Response


class CLIRef:
    """
    A CLI reference to a model

    The cli program must conform to the Meillionen cli interface
    """

    def __init__(self, path):
        self.command = sh.Command(path)

    def get_interface(self) -> ModuleInterface:
        interface = self.command('interface')
        module = ModuleInterface.deserialize(interface.stdout)
        return module

    def run(self, mr: MethodRequest):
        builder = flatbuffers.Builder()
        # need to prefix the message by the size
        # so that calling program knows how much of
        # stdin is needed to interpret the MethodRequest
        builder.FinishSizePrefixed(mr.serialize(builder))
        msg = builder.Output()
        self.command('run', _in=bytes(msg))


class DockerImageRef:
    """
    A Docker Image reference to a model

    The docker image must have an entrypoint of a Meillionen cli program and
    have a working directory of `/code` in order to work
    """

    def __init__(self, image_name: str):
        # FIXME: support additional docker volume mounts for data that is not
        #  available in the directory the model is currently running in
        self.image_name = image_name
        curdir = os.getcwd()
        self.command(f'docker run -it --rm -v {curdir}:/code {image_name}')

    def get_interface(self) -> ModuleInterface:
        self.command('interface')
        module = ModuleInterface.deserialize(interface.stdout)
        return module

    def run(self, mr: MethodRequest):
        builder = flatbuffers.Builder()
        builder.FinishSizePrefixed(mr.serialize(builder))
        msg = builder.Output()
        self.command('run', _in=bytes(msg))


class ServerRef:
    """
    A server reference to a model

    This allows running a model in the same Python process.

    Mostly suited for interactive experimentation and model testing.
    """
    def __init__(self, server: Server):
        self.server = server

    def get_interface(self) -> ModuleInterface:
        return self.server.module

    def run(self, mr: MethodRequest):
        self.server.run(mr)


class Client:
    """
    A client to execute and inspect Meillionen models
    """
    def __init__(self, module_ref, settings=None):
        self.module_ref = module_ref
        self.module: ModuleInterface = module_ref.get_interface()
        self.settings = settings

    def run_simple(self, mr: MethodRequest):
        self.module_ref.run(mr)
        method = self.module.get_method(mr)
        resources = mr.kwargs
        handlers = get_handlers(resources=resources, schemas=method.args)
        return Response(handlers=handlers, resources=resources)

    def run(self, class_name, method_name, resource_payloads, partition: Optional[Partition]=None):
        """
        Runs a particular class method on the module

        :param class_name: the name of the class
        :param method_name: the name of the method to call
        :param resource_payloads: dict of argument names to resource payloads
        :param partition: an optional partition which describes how to build nested paths,
          to add key columns to tabular data or to add dimensions to tensor data
        """
        mr = MethodRequest.from_partial(
            class_name=class_name,
            method_name=method_name,
            resource_payloads=resource_payloads,
            partition=partition,
            settings=self.settings
        )
        return self.run_simple(mr)

    def save(self, mra: MethodRequestArg, resource, data, partition=None):
        """
        Saves a resource

        :param mra: metadata needed to locate the schema of a method argument
        :param resource: the resource metadata describing how and where to save the data to
        :param data: the data to save
        :param partition: the optional partition information. Useful for evaluating a model at
          many different parameter combinations
        """
        if hasattr(resource, 'complete'):
            resource = resource.complete(settings=self.settings, mra=mra, partition=partition)
        method = self.module.get_method(mra)
        schema = method.args[mra.arg_name]
        handler = get_handler(resource_payload=resource, schema=schema)
        handler.save(resource, data)
        return resource


import functools
import io
import json
import os.path
import pathlib
from typing import Union, Dict, List, Any

import flatbuffers
import netCDF4
import numpy as np
import pandas as pd
import pyarrow as pa
import xarray as xr
from landlab.io import read_esri_ascii, write_esri_ascii

from abc import abstractmethod
from typing import Optional, Protocol, Type, TypeVar
from . import _Schema as s
from .mutability import Mutability
from .base import field_to_bytesio
from ..exceptions import HandlerNotFound, ExtensionHandlerNotFound
from .resource import get_resource_payload_class, Feather, Parquet, NetCDF, OtherFile, ResourcePayloadable


def _mkdir_p(path):
    p = pathlib.Path(path)
    p.parent.mkdir(parents=True, exist_ok=True)


_SP = TypeVar('_SP')


class SchemaPayloadable(Protocol):
    """
    A schema payload describes the shape of data a method argument expects
    """

    name: str

    @classmethod
    @abstractmethod
    def deserialize(cls: Type[_SP], buffer) -> _SP:
        """
        Builds the schema payload from a buffer

        :param buffer: the buffer to load from
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize(self, builder: flatbuffers.Builder):
        """
        Serializes the schema payload into a flatbuffer builder

        :param builder: the flatbuffer builder
        """
        raise NotImplementedError()


class DataFrameSchemaPayload(SchemaPayloadable):
    name = 'meillionen::schema::DataFrameSchema'

    def __init__(self, arrow_schema: pa.Schema):
        self.arrow_schema = arrow_schema

    @classmethod
    def deserialize(cls, buffer: io.BytesIO):
        schema = pa.ipc.read_schema(buffer)
        return cls(arrow_schema=schema)

    def serialize(self, builder: flatbuffers.Builder):
        buf: pa.Buffer = self.arrow_schema.serialize()
        off = builder.CreateByteVector(buf.to_pybytes())
        return off


class TensorSchemaPayload(SchemaPayloadable):
    name = 'meillionen::schema::TensorSchema'

    def __init__(self, data_type: str, dimensions: List[str]):
        self.data_type = data_type
        self.dimensions = dimensions

    @classmethod
    def deserialize(cls, buffer: io.BytesIO):
        data = json.load(buffer)
        return cls(data_type=data['data_type'], dimensions=data['dimensions'])

    def serialize(self, builder: flatbuffers.Builder):
        data = json.dumps({'data_type': self.data_type, 'dimensions': self.dimensions}).encode('utf-8')
        return builder.CreateByteVector(data)


class SchemalessPayload(SchemaPayloadable):
    name = 'meillionen::schema::Schemaless'

    @classmethod
    def deserialize(cls, buffer: io.BytesIO):
        return cls()

    def serialize(self, builder: flatbuffers.Builder):
        data = json.dumps('').encode('utf-8')
        return builder.CreateByteVector(data)


_SCHEMA_CLASSES = {
    s.name: s for s in
    [
        DataFrameSchemaPayload,
        TensorSchemaPayload,
        SchemalessPayload
    ]
}


def clear_schema_payload_classes():
    _SCHEMA_CLASSES.clear()


def register_schema_payload_class(cls):
    _SCHEMA_CLASSES[cls.name] = cls


def get_schema_payload_class(name):
    return _SCHEMA_CLASSES[name]


class _Schema(s._Schema):
    PAYLOAD_OFFSET = 10

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = cls()
        x.Init(buf, n + offset)
        return x

    def PayloadAsBytesIO(self):
        return field_to_bytesio(tab=self._tab, field_offset=self.PAYLOAD_OFFSET)


class Schema:
    def __init__(self, name, schema, resource_classes, mutability: Mutability):
        self.name = name
        self.schema = schema
        self.resource_classes = resource_classes
        self.mutability = mutability

    def _serialize_resource_names(self, builder: flatbuffers.Builder, resource_classes: List[Any]):
        n = len(resource_classes)
        offsets = [builder.CreateString(resource_classes[i].name) for i in range(n)]
        s.StartResourceNamesVector(builder, n)
        for i in range(n):
            builder.PrependUOffsetTRelative(offsets[i])
        return builder.EndVector()

    def serialize(self, builder: flatbuffers.Builder):
        name_off = builder.CreateString(self.name)
        type_name_off = builder.CreateString(self.schema.name)
        payload_off = self.schema.serialize(builder)
        resource_names_off = self._serialize_resource_names(builder, self.resource_classes)
        s.Start(builder)
        s.AddName(builder, name_off)
        s.AddTypeName(builder, type_name_off)
        s.AddMutability(builder, self.mutability.serialize())
        s.AddPayload(builder, payload_off)
        s.AddResourceNames(builder, resource_names_off)
        schema_off = s.End(builder)
        return schema_off

    @classmethod
    def from_class(cls, schema: _Schema):
        name = schema.Name().decode('utf-8')
        mutability = Mutability(schema.Mutability())
        payload_class = get_schema_payload_class(schema.TypeName().decode('utf-8'))
        payload = payload_class.deserialize(schema.PayloadAsBytesIO())
        resource_classes = []
        for i in range(schema.ResourceNamesLength()):
            resource_name = schema.ResourceNames(i).decode('utf-8')
            resource_class = get_resource_payload_class(resource_name)
            resource_classes.append(resource_class)
        return cls(name=name, schema=payload, resource_classes=resource_classes, mutability=mutability)


class SchemaProxy:
    @property
    def name(self):
        return self.schema.name

    @property
    def mutability(self):
        return self.schema.mutability

    @property
    def resource_classes(self):
        return self.schema.resource_classes


_H = TypeVar('_H')


class Handlable(Protocol):
    @classmethod
    @abstractmethod
    def from_schema(cls: Type[_H], schema: Schema) -> _H:
        """
        Builds the handler from the schema

        :param schema: the schema to build the handler with
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize(self, builder: flatbuffers.Builder):
        """
        Serialize a handler into a flatbuffer builder using only schema information
        """
        raise NotImplementedError()

    @abstractmethod
    def load(self, resource):
        """
        Loads a resource into memory

        :param resource: the resource to load
        """
        raise NotImplementedError()

    @abstractmethod
    def save(self, resource, data):
        """
        Save data using the metadata provided by the resource

        :param resource: the metadata desribing where and how to save the data
        :param data: the data to save
        """
        raise NotImplementedError()


class PandasHandler(SchemaProxy, Handlable):
    RESOURCE_CLASSES = [Feather, Parquet]

    def __init__(self, name: str, s: pa.Schema, mutability: Mutability):
        self.schema = Schema(
            name=name,
            schema=DataFrameSchemaPayload(s),
            resource_classes=self.RESOURCE_CLASSES,
            mutability=mutability)

    @classmethod
    def from_schema(cls, schema: Schema):
        return cls(
            name=schema.name,
            s=schema.schema.arrow_schema,
            mutability=schema.mutability
        )

    def serialize(self, builder: flatbuffers.Builder):
        return self.schema.serialize(builder)

    @functools.singledispatchmethod
    def load(self, resource):
        raise NotImplemented()

    @load.register
    def _load(self, resource: Feather):
        return pd.read_feather(resource.path)

    @load.register
    def _load(self, resource: Parquet):
        return pd.read_parquet(resource.path)

    @functools.singledispatchmethod
    def save(self, resource, data):
        raise NotImplemented()

    @save.register
    def _save(self, resource: Feather, data: pd.DataFrame):
        _mkdir_p(resource.path)
        return data.to_feather(resource.path)

    @save.register
    def _save(self, resource: Parquet, data: pd.DataFrame):
        _mkdir_p(resource.path)
        return data.to_parquet(resource.path)


class NetCDFHandler(SchemaProxy, Handlable):
    RESOURCE_CLASSES = [NetCDF]

    def __init__(self, name: str, data_type: str, dimensions: List[str], mutability: Mutability):
        ts = TensorSchemaPayload(data_type=data_type, dimensions=dimensions)
        self.schema = Schema(
            name=name,
            schema=ts,
            resource_classes=self.RESOURCE_CLASSES,
            mutability=mutability)

    @classmethod
    def from_schema(cls, schema: Schema):
        return cls(
            name=schema.name,
            data_type=schema.schema.data_type,
            dimensions=schema.schema.dimensions,
            mutability=schema.mutability)

    @property
    def data_type(self):
        return self.schema.schema.data_type

    @property
    def dimensions(self):
        return self.schema.schema.dimensions

    def serialize(self, builder: flatbuffers.Builder):
        return self.schema.serialize(builder)

    @functools.singledispatchmethod
    def load(self, resource):
        raise NotImplemented()

    @load.register
    def _load(self, resource: NetCDF):
        ds = netCDF4.Dataset(resource.path)
        try:
            return ds[resource.variable][:]
        finally:
            if ds.isopen():
                ds.close()

    @functools.singledispatchmethod
    def save(self, resource, data):
        raise NotImplemented()

    @save.register
    def _save(self, resource: NetCDF, data: xr.DataArray):
        data = data.transpose(self.dimensions)
        ds, variable = _netcdf_create_variable(self.schema.schema, resource=resource, dimensions=data.dims)
        try:
            variable[:] = data
        finally:
            if ds.isopen():
                ds.close()


def _netcdf_create_variable(schema, resource, dimensions):
    dimnames = schema.dimensions
    _mkdir_p(resource.path)
    dataset = netCDF4.Dataset(resource.path, mode='w')
    for dim in dimnames:
        size = dimensions[dim]
        print((dim, size))
        dataset.createDimension(dim, size)
    variable = dataset.createVariable(resource.variable, 'f4', dimnames)
    return dataset, variable


class NetCDFSliceHandler(SchemaProxy, Handlable):
    RESOURCE_CLASSES = [NetCDF]

    def __init__(self, name: str, data_type: str, dimensions: List[str], mutability: Mutability):
        self.schema = Schema(
            name=name,
            schema=TensorSchemaPayload(data_type=data_type, dimensions=dimensions),
            resource_classes=self.RESOURCE_CLASSES,
            mutability=mutability
        )

    @classmethod
    def from_schema(cls, schema: Schema):
        return cls(
            name=schema.name,
            data_type=schema.schema.data_type,
            dimensions=schema.schema.dimensions,
            mutability=schema.mutability
        )

    def serialize(self, builder: flatbuffers.Builder):
        return self.schema.serialize(builder)

    @functools.singledispatchmethod
    def load(self, resource):
        raise NotImplemented()

    @load.register
    def _load(self, resource: NetCDF):
        return NetCDFSliceLoader(resource=resource)

    @functools.singledispatchmethod
    def save(self, resource, data):
        raise NotImplemented()

    @save.register
    def _save(self, resource: NetCDF, data):
        return NetCDFSliceSaver(schema=self.schema, resource=resource, dimensions=data)


NDSlice = Dict[str, Union[int, slice]]


class NetCDFSliceLoader:
    def __init__(self, resource: NetCDF):
        self.resource = resource
        self.dataset = netCDF4.Dataset(resource.path, mode='r')
        self.variable = self.dataset[resource.variable]

    def get(self, slices: NDSlice):
        vdims = self.variable.dimensions
        xs = tuple(slices[d] if d in slices else slice(None) for d in vdims)
        return self.variable[xs]


class NetCDFSliceSaver:
    def __init__(self, schema, resource: NetCDF, dimensions):
        self.schema = schema
        self.dataset, self.variable = _netcdf_create_variable(
            schema=schema.schema,
            resource=resource,
            dimensions=dimensions)

    def set(self, slices: Dict[str, Union[int, slice]], array: xr.DataArray):
        """
        Save an array to a slice of a NetCDF variable

        :param array: A dimensioned array
        :param slices: Where to save the dimensioned array to. Dimensions of array passed in
          and NetCDF variable names must match
        """
        vdims = self.variable.dimensions
        xs = tuple(slices[d] if d in slices else slice(None) for d in vdims)
        vdims_remaining = [v for v in vdims if v not in slices.keys()]
        self.variable[xs] = array.transpose(*vdims_remaining)

    def _close(self):
        self.dataset.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._close()


class LandLabGridHandler(SchemaProxy, Handlable):
    RESOURCE_CLASSES = [OtherFile]

    def __init__(self, name: str, data_type: str, mutability: Mutability):
        dimensions = ['x', 'y']
        ts = TensorSchemaPayload(data_type=data_type, dimensions=dimensions)
        self.schema = Schema(name=name, schema=ts, resource_classes=self.RESOURCE_CLASSES, mutability=mutability)

    @classmethod
    def from_schema(cls, schema: Schema):
        return cls(
            name=schema.name,
            data_type=schema.schema.data_type,
            mutability=schema.mutability)

    def serialize(self, builder: flatbuffers.Builder):
        return self.schema.serialize(builder)

    @functools.singledispatchmethod
    def load(self, resource):
        raise NotImplemented()

    @load.register
    def _load(self, resource: OtherFile):
        mg, z = read_esri_ascii(resource.path, name='topographic__elevation')
        mg.status_at_node[mg.nodes_at_right_edge] = mg.BC_NODE_IS_FIXED_VALUE
        mg.status_at_node[np.isclose(z, -9999)] = mg.BC_NODE_IS_CLOSED
        return mg

    @functools.singledispatchmethod
    def save(self, resource, data):
        raise NotImplemented()

    @save.register
    def _save(self, resource: OtherFile, data):
        write_esri_ascii(resource.path, data)


class DirHandler(SchemaProxy, Handlable):
    RESOURCE_CLASSES = [OtherFile]

    def __init__(self, name: str, mutability: Mutability):
        self.schema = Schema(
            name=name,
            schema=SchemalessPayload(),
            resource_classes=self.RESOURCE_CLASSES,
            mutability=mutability)

    @classmethod
    def from_schema(cls, schema: Schema):
        return cls(
            name=schema.name,
            mutability=schema.mutability)

    def serialize(self, builder: flatbuffers.Builder):
        return self.schema.serialize(builder)

    def load(self, resource):
        raise NotImplemented()

    @functools.singledispatchmethod
    def save(self, resource):
        raise NotImplemented()

    @save.register
    def _save(self, resource: OtherFile):
        path = resource.path
        p = pathlib.Path(path)
        p.mkdir(parents=True, exist_ok=True)
        return path


class PassthroughHandlerMapper:
    def __init__(self, handler_cls):
        self.handler_cls = handler_cls

    def to_handler(self, resource, schema):
        return self.handler_cls.from_schema(schema)


class FileExtensionHandlerMapper:
    """
    A handler mapper that inspects a resources file extension to determine what handler
    should wrap a particular schema
    """
    def __init__(self):
        self.ext_map = {}

    def register_ext(self, ext: str, handler_cls):
        self.ext_map[ext] = handler_cls

    def to_handler(self, resource: OtherFile, schema: Schema):
        ext = os.path.splitext(resource.path)[1]
        try:
            handler_cls = self.ext_map[ext]
        except KeyError as e:
            raise ExtensionHandlerNotFound(ext) from e
        return handler_cls.from_schema(schema)

FILE_EXT_HANDLER_MAPPER = FileExtensionHandlerMapper()
FILE_EXT_HANDLER_MAPPER.register_ext('.asc', LandLabGridHandler)
FILE_EXT_HANDLER_MAPPER.register_ext('', DirHandler)

_RESOURCE_HANDLER_MAPPER = {
    OtherFile: FILE_EXT_HANDLER_MAPPER,
    Feather: PassthroughHandlerMapper(PandasHandler),
    Parquet: PassthroughHandlerMapper(PandasHandler),
    NetCDF: PassthroughHandlerMapper(NetCDFHandler),
}


def clear_handler_classes():
    _RESOURCE_HANDLER_MAPPER.clear()


def register_resource_handler_mapper(resource_cls, handler_mapper):
    _RESOURCE_HANDLER_MAPPER[resource_cls] = handler_mapper


def all_resource_handler_mappers(name):
    return _RESOURCE_HANDLER_MAPPER


def get_handler(resource_payload, schema: Schema):
    """
    Retrieve a handler given a resource payload and a schema
    """
    handler_mapper = _RESOURCE_HANDLER_MAPPER[type(resource_payload)]
    return handler_mapper.to_handler(resource_payload, schema)


def get_handlers(resources, schemas: Dict[str, Schema]):
    handlers = {}
    for name, schema in schemas.items():
        resource = resources[name]
        handler = get_handler(resource, schema)
        handlers[name] = handler
    return handlers


from meillionen.client import ClientFunctionModel


class PyMTFunctionModel:
    def __init__(self):
        self.sources = {}
        self.sinks = {}
        self.model: ClientFunctionModel = None
        self.partition = None

    def initialize(self, model):
        self.model = model

    @classmethod
    def setup(cls):
        return None

    def get_input_var_names(self):
        return self.model.source_names

    def get_input_var_type(self, name):
        return self.model.source_schema(name)

    def get_output_var_names(self):
        return self.model.sink_names

    def get_output_var_type(self, name):
        return self.model.sink_schema(name)

    def set_partition(self, partition):
        self.partition = partition

    def set_value(self, source_name, source):
        self.sources[source_name] = source

    def get_value(self, sink_name):
        return self.sinks[sink_name]

    def update(self):
        self.sinks = self.model.run(sources=self.sources, partition=self.partition)

    def finalize(self):
        self.sinks = {}


from typing import Dict, Any

import flatbuffers
import io


def deserialize_to_dict(constructor, getter, n):
    xs = {}
    for i in range(n):
        x = constructor(getter(i))
        xs[x.name] = x
    return xs


def serialize_dict(builder: flatbuffers.Builder, vector_builder, xs: Dict[str, Any]):
    xs_list_off = []
    for x in xs.values():
        xs_off = x.serialize(builder)
        xs_list_off.append(xs_off)

    vector_builder(builder, len(xs))
    for off in xs_list_off:
        builder.PrependUOffsetTRelative(off)
    return builder.EndVector()


def serialize_list(builder: flatbuffers.Builder, vector_builder, xs):
    xs_list_off = []
    for x in xs:
        xs_off = x.serialize(builder)
        xs_list_off.append(xs_off)

    vector_builder(builder, len(xs))
    for off in xs_list_off:
        builder.PrependUOffsetTRelative(off)
    return builder.EndVector()


def field_to_bytesio(tab, field_offset):
    o = flatbuffers.number_types.UOffsetTFlags.py_type(tab.Offset(field_offset))
    if o == 0:
        return None
    offset = tab.Vector(o)
    length = tab.VectorLen(o)
    bo = io.BytesIO(memoryview(tab.Bytes)[offset:(offset + length)])
    return bo


class FlatbufferMixin:
    @classmethod
    def get_root_as(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = cls()
        x.Init(buf, n + offset)
        return x

    def _get_resource(self, j, offset, klass):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(offset))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            obj = klass()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None


class MethodRequestArg:
    def __init__(self, class_name: str, method_name: str, arg_name: str):
        self.class_name = class_name
        self.method_name = method_name
        self.arg_name = arg_name


from typing import Dict, List, Union, Any

import flatbuffers

from . import _ClassInterface as ci
from .method_interface import MethodInterface, _MethodInterface
from .base import serialize_list, serialize_dict, FlatbufferMixin
from meillionen.exceptions import MethodNotFound


class _ClassInterface(ci._ClassInterface, FlatbufferMixin):
    METHOD_OFFSET = 8

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        return cls.get_root_as(buf, offset)

    def Methods(self, j):
        return self._get_resource(j, self.METHOD_OFFSET, _MethodInterface)


class ClassInterface:
    """
    A class interface describes the methods available in a model along with the model's name
    """
    def __init__(self, name, methods: Union[List[MethodInterface], Dict[str, MethodInterface]]):
        self._name = name
        self._methods = {m.name: m for m in methods} if not hasattr(methods, 'values') else methods

    @property
    def name(self):
        """
        :return: the name of the model
        """
        return self._name

    @classmethod
    def from_interface(cls, interface: _ClassInterface):
        """
        Builds a :class:`~ClassInterface` from a flatbuffer :class:`~_ClassInterface`
        """
        name = interface.Name().decode('utf-8')
        methods = {}
        for i in range(interface.MethodsLength()):
            method = MethodInterface.from_interface(interface.Methods(i))
            methods[method.name] = method
        return cls(name=name, methods=methods)

    @classmethod
    def deserialize(cls, buffer):
        """
        Builds a :class:`~ClassInterface` from a buffer
        """
        interface = _ClassInterface.GetRootAs(buffer, 0)
        return cls.from_interface(interface)

    def serialize(self, builder: flatbuffers.Builder):
        """
        Serialized a class interface into a flatbuffer builder
        """
        name_off = builder.CreateString(self._name)
        methods_off = serialize_dict(
            builder=builder,
            vector_builder=ci.StartMethodsVector,
            xs=self._methods
        )

        ci.Start(builder)
        ci.AddName(builder, name_off)
        ci.AddMethods(builder, methods_off)
        return ci.End(builder)

    def get_method(self, name: str) -> MethodInterface:
        """
        Retrieve a method from a class given its name
        """
        try:
            return self._methods[name]
        except KeyError as e:
            raise MethodNotFound() from e



import flatbuffers
import json
import numpy as np
import os.path

from typing import Optional, Protocol, TypeVar, Type
from . import _Resource as r
from .base import field_to_bytesio, MethodRequestArg
from meillionen.exceptions import ResourceNotFound
from meillionen.settings import Settings, Partition


class _Resource(r._Resource):
    PAYLOAD_OFFSET = 8

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = cls()
        x.Init(buf, n + offset)
        return x

    def PayloadAsBytesIO(self):
        return field_to_bytesio(tab=self._tab, field_offset=self.PAYLOAD_OFFSET)


class Resource:
    """A serializable resource. Should only be used internally."""
    def __init__(self, name: str, resource_payload):
        self.name = name
        self.resource_payload = resource_payload

    @classmethod
    def from_class(cls, resource: _Resource):
        name = resource.Name().decode('utf-8')
        type_name = resource.TypeName().decode('utf-8')
        payload_class = get_resource_payload_class(type_name)
        resource_payload = payload_class.deserialize(resource.PayloadAsBytesIO())
        return cls(name=name, resource_payload=resource_payload)

    def serialize(self, builder: flatbuffers.Builder):
        name_off = builder.CreateString(self.name)
        type_name_off = builder.CreateString(self.resource_payload.name)
        payload_off = self.resource_payload.serialize(builder)
        r.Start(builder)
        r.AddName(builder, name_off)
        r.AddTypeName(builder, type_name_off)
        r.AddPayload(builder, payload_off)
        return r.End(builder)


class PartialResource:
    def __init__(self, resource_payload_class, kwargs):
        self.resource_payload_class = resource_payload_class
        self.kwargs = kwargs

    def complete(self, settings, mra: MethodRequestArg, partition=None):
        return self.resource_payload_class.from_kwargs(
            kwargs=self.kwargs,
            settings=settings,
            mra=mra,
            partition=partition)


def build_path(settings, mra: MethodRequestArg, ext, partition=None):
    base_path = os.path.join(settings.base_path, mra.class_name, mra.method_name, mra.arg_name)
    if partition:
        path = os.path.join(base_path, partition.to_path(), f'data{ext}')
    else:
        path = f'{base_path}{ext}'
    return path


_R = TypeVar("_R")


class ResourcePayloadable(Protocol):
    """
    The interface of a resource
    """
    @classmethod
    def from_kwargs(cls: Type[_R], kwargs, settings: Settings, mra: MethodRequestArg, partition: Optional[Partition]) -> _R:
        """
        Builds a resource payload

        :param kwargs: the arguments to pass to the classes initializer
        :param settings: the common settings used to complete the partial resource payload
        :param mra: the metadata describing which argument of a method is being built
        :param partition: additional context used to build the resource for grid searches and sensitivity analysis
        """
        raise NotImplementedError()

    @classmethod
    def deserialize(cls: Type[_R], buffer) -> _R:
        """
        Builds the resource payload from a buffer
        """
        raise NotImplementedError()

    def serialize(self, builder: flatbuffers.Builder):
        """
        Serializes the resource payload into the flatbuffer builder

        :param builder: the flatbuffer builder
        """
        raise NotImplementedError()


class OtherFile(ResourcePayloadable):
    """A file payload resource payload

    Together with a name a full resource can be created"""
    name = 'meillionen::resource::OtherFile'

    def __init__(self, path):
        self.path = path

    @classmethod
    def partial(cls, ext):
        return PartialResource(cls, {'ext': ext})

    @classmethod
    def from_kwargs(cls, kwargs, settings, mra: MethodRequestArg, partition):
        path = build_path(
            settings=settings,
            mra=mra,
            partition=partition,
            ext=kwargs['ext']
        )
        return cls(path=path)

    @classmethod
    def deserialize(cls, buffer):
        path = json.load(buffer)['path']
        return cls(path=path)

    def serialize(self, builder: flatbuffers.Builder):
        return builder.CreateByteVector(json.dumps({'path': self.path}).encode('utf-8'))


class Parquet(ResourcePayloadable):
    """A Parquet file resource payload"""

    name = 'meillionen::resource::Parquet'
    ext = '.parquet'

    def __init__(self, path):
        self.path = path

    @classmethod
    def partial(cls):
        return PartialResource(cls, {})

    @classmethod
    def from_kwargs(cls, kwargs, settings, mra: MethodRequestArg, partition=None):
        path = build_path(
            settings=settings,
            mra=mra,
            partition=partition,
            ext=cls.ext
        )
        return cls(path=path)

    @classmethod
    def deserialize(cls, buffer):
        path = json.load(buffer)['path']
        return cls(path=path)

    def serialize(self, builder: flatbuffers.Builder):
        data = json.dumps({'path': self.path}).encode('utf-8')
        return builder.CreateByteVector(data)


class Feather(ResourcePayloadable):
    """A Feather file resource payload"""

    name = 'meillionen::resource::Feather'
    ext = '.feather'

    def __init__(self, path):
        self.path = path

    @classmethod
    def partial(cls):
        return PartialResource(cls, {})

    @classmethod
    def from_kwargs(cls, kwargs, settings, mra: MethodRequestArg, partition=None):
        path = build_path(
            settings=settings,
            mra=mra,
            ext=cls.ext,
            partition=partition
        )
        return cls(path=path)

    @classmethod
    def deserialize(cls, buffer):
        path = json.load(buffer)['path']
        return cls(path=path)

    def serialize(self, builder: flatbuffers.Builder):
        data = np.frombuffer(json.dumps({'path': self.path}).encode('utf-8'), dtype=np.uint8)
        return builder.CreateNumpyVector(data)


class NetCDF(ResourcePayloadable):
    """A NetCDF file resource payload"""

    name = 'meillionen::resource::NetCDF'
    ext = '.nc'

    def __init__(self, path, variable):
        self.path = path
        self.variable = variable

    @classmethod
    def partial(cls, variable):
        return PartialResource(cls, {'variable': variable})

    @classmethod
    def from_kwargs(cls, kwargs, settings, mra: MethodRequestArg, partition=None):
        path = build_path(
            settings=settings,
            mra=mra,
            ext=cls.ext,
            partition=partition
        )
        return cls(path=path, variable=kwargs['variable'])

    @classmethod
    def deserialize(cls, buffer):
        data = json.load(buffer)
        path = data['path']
        variable = data['variable']
        return cls(path=path, variable=variable)

    def serialize(self, builder: flatbuffers.Builder):
        data = json.dumps({'path': self.path, 'variable': self.variable}).encode('utf-8')
        return builder.CreateByteVector(data)


_RESOURCE_CLASSES = {
    r.name: r for r in
    [
        Feather,
        NetCDF,
        Parquet,
        OtherFile
    ]
}


def deserialize_resource_payload(resource: _Resource):
    type_name = resource.TypeName().decode('utf-8')
    buffer = resource.PayloadAsBytesIO()
    try:
        return get_resource_payload_class(type_name).deserialize(buffer)
    except KeyError as e:
        raise ResourceNotFound() from e


def clear_resource_payload_classes():
    _RESOURCE_CLASSES.clear()


def register_resource_payload_class(cls):
    _RESOURCE_CLASSES[cls.name] = cls


def get_resource_payload_class(name):
    return _RESOURCE_CLASSES[name]


import io

import flatbuffers
from meillionen.interface.base import MethodRequestArg

from ..settings import Settings
from . import _MethodRequest as mr
from .resource import deserialize_resource_payload, Resource
from .base import FlatbufferMixin


class _MethodRequest(mr._MethodRequest, FlatbufferMixin):
    ARGS_OFFSET = 8

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        return cls.get_root_as(buf, offset)

    # _MethodRequest
    def Args(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(self.ARGS_OFFSET))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from meillionen.interface.resource import _Resource
            obj = _Resource()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None


class MethodRequest:
    """
    A request to call a model method
    """
    def __init__(self, class_name, method_name, kwargs):
        self.class_name = class_name
        self.method_name = method_name
        self.kwargs = kwargs

    @classmethod
    def from_partial(cls, settings: Settings, class_name: str, method_name: str, resource_payloads, partition):
        """
        settings:
        """
        kwargs = {}
        for name, resource_payload in resource_payloads.items():
            mra = MethodRequestArg(class_name=class_name, method_name=method_name, arg_name=name)
            if hasattr(resource_payload, 'complete'):
                kwargs[name] = resource_payload.complete(settings=settings, mra=mra, partition=partition)
            else:
                kwargs[name] = resource_payload
        return cls(class_name=class_name, method_name=method_name, kwargs=kwargs)

    @staticmethod
    def _serialize_resources(builder: flatbuffers.Builder, resources):
        offsets = []
        for name, resource_payload in resources.items():
            resource = Resource(name=name, resource_payload=resource_payload)
            offsets.append(resource.serialize(builder))
        mr.StartArgsVector(builder, len(offsets))
        for offset in offsets:
            builder.PrependUOffsetTRelative(offset)
        return builder.EndVector()

    @staticmethod
    def _deserialize_resources(getter, n):
        resources = {}
        for i in range(n):
            resource = getter(i)
            name = resource.Name().decode('utf-8')
            resource_payload = deserialize_resource_payload(resource)
            resources[name] = resource_payload
        return resources

    @classmethod
    def deserialize(cls, buffer: io.BytesIO):
        req = _MethodRequest.GetRootAs(buffer, 0)
        class_name = req.ClassName().decode('utf-8')
        method_name = req.MethodName().decode('utf-8')
        kwargs = cls._deserialize_resources(getter=req.Args, n=req.ArgsLength())
        return cls(
            class_name=class_name,
            method_name=method_name,
            kwargs=kwargs
        )

    def serialize(self, builder: flatbuffers.Builder):
        """
        Serialize a method request into a flatbuffer builder

        :param builder: the flatbuffer builder
        """
        cls_off = builder.CreateString(self.class_name)
        method_off = builder.CreateString(self.method_name)
        kwargs_off = self._serialize_resources(builder, self.kwargs)
        mr.Start(builder)
        mr.AddClassName(builder, cls_off)
        mr.AddMethodName(builder, method_off)
        mr.AddArgs(builder, kwargs_off)
        return mr.End(builder)

    def get_arg(self, name):
        """
        Get the path metadata for a argument

        :param name: the name of the argument
        """
        return MethodRequestArg(class_name=self.class_name, method_name=self.method_name, arg_name=self.kwargs[name])


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: interface

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class _ClassInterface(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _ClassInterface()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAs_ClassInterface(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # _ClassInterface
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # _ClassInterface
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # _ClassInterface
    def TypeName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # _ClassInterface
    def Methods(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from meillionen.interface._FunctionInterface import _FunctionInterface
            obj = _FunctionInterface()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # _ClassInterface
    def MethodsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # _ClassInterface
    def MethodsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def Start(builder): builder.StartObject(3)
def _ClassInterfaceStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddName(builder, name): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def _ClassInterfaceAddName(builder, name):
    """This method is deprecated. Please switch to AddName."""
    return AddName(builder, name)
def AddTypeName(builder, typeName): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(typeName), 0)
def _ClassInterfaceAddTypeName(builder, typeName):
    """This method is deprecated. Please switch to AddTypeName."""
    return AddTypeName(builder, typeName)
def AddMethods(builder, methods): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(methods), 0)
def _ClassInterfaceAddMethods(builder, methods):
    """This method is deprecated. Please switch to AddMethods."""
    return AddMethods(builder, methods)
def StartMethodsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def _ClassInterfaceStartMethodsVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartMethodsVector(builder, numElems)
def End(builder): return builder.EndObject()
def _ClassInterfaceEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


from .class_interface import ClassInterface
from .method_interface import MethodInterface
from .method_request import MethodRequest
from .module_interface import ModuleInterface
from .mutability import Mutability



import enum

from ._Mutability import _Mutability


class Mutability(enum.Enum):
    """
    The mutability of a schema.

    A readable schema is an input argument to a :class:`~meillionen.interface.method_interface.MethodInterface`

    A writable schema is an output argument to a :class:`~meillionen.interface.method_interface.MethodInterface`
    """
    read = _Mutability.read
    write = _Mutability.write

    def serialize(self):
        """
        Converts the mutability enum into a flatbuffer serializable format
        """
        return self.value

    @classmethod
    def deserialize(cls, value: int):
        """
        Turns the mutability int into an enum
        """
        return cls(value)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: interface

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class _MethodRequest(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _MethodRequest()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAs_MethodRequest(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # _MethodRequest
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # _MethodRequest
    def ClassName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # _MethodRequest
    def MethodName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # _MethodRequest
    def Args(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from meillionen.interface._Resource import _Resource
            obj = _Resource()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # _MethodRequest
    def ArgsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # _MethodRequest
    def ArgsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def Start(builder): builder.StartObject(3)
def _MethodRequestStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddClassName(builder, className): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(className), 0)
def _MethodRequestAddClassName(builder, className):
    """This method is deprecated. Please switch to AddClassName."""
    return AddClassName(builder, className)
def AddMethodName(builder, methodName): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(methodName), 0)
def _MethodRequestAddMethodName(builder, methodName):
    """This method is deprecated. Please switch to AddMethodName."""
    return AddMethodName(builder, methodName)
def AddArgs(builder, args): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(args), 0)
def _MethodRequestAddArgs(builder, args):
    """This method is deprecated. Please switch to AddArgs."""
    return AddArgs(builder, args)
def StartArgsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def _MethodRequestStartArgsVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartArgsVector(builder, numElems)
def End(builder): return builder.EndObject()
def _MethodRequestEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: interface

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class _Resource(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _Resource()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAs_Resource(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # _Resource
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # _Resource
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # _Resource
    def TypeName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # _Resource
    def Payload(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # _Resource
    def PayloadAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # _Resource
    def PayloadLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # _Resource
    def PayloadIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def Start(builder): builder.StartObject(3)
def _ResourceStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddName(builder, name): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def _ResourceAddName(builder, name):
    """This method is deprecated. Please switch to AddName."""
    return AddName(builder, name)
def AddTypeName(builder, typeName): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(typeName), 0)
def _ResourceAddTypeName(builder, typeName):
    """This method is deprecated. Please switch to AddTypeName."""
    return AddTypeName(builder, typeName)
def AddPayload(builder, payload): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(payload), 0)
def _ResourceAddPayload(builder, payload):
    """This method is deprecated. Please switch to AddPayload."""
    return AddPayload(builder, payload)
def StartPayloadVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def _ResourceStartPayloadVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartPayloadVector(builder, numElems)
def End(builder): return builder.EndObject()
def _ResourceEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


import os.path
import pyarrow as pa
import pyarrow.dataset as pad

from typing import Any, Dict
from pydantic import BaseModel


class Settings(BaseModel):
    """
    Settings for a module.

    Currently used only to turn partial resource payloads into resource payloads. In the future
    it may have information about logging and how to access particular common resources (such as
    databases).
    """
    base_path: str

    def trial(self, name: str):
        """
        Used to build 'trial' settings objects which extend the base path of the settings object.
        """
        return self.__class__(base_path=os.path.join(self.base_path, name))


class Partitioning:
    """
    Describes additional data to be used to build up partial resource payload
    objects.

    Commonly used when broadcasting a model over a grid with different resources
    available at each grid point. Also useful for sensitivity analysis.
    """
    def __init__(self, schema: pa.Schema):
        self.schema = schema

    @property
    def names(self):
        return self.schema.names

    def to_arrow(self):
        """
        Convert the partition into a `pyarrow` partition so that final results can
        be loaded as partitioned parquet files (which are supported by pyarrow, spark,
        datafusion and other libraries)
        """
        return pad.partitioning(self.schema)

    def complete(self, **kwargs):
        """
        Build a partition object with the data filled in for each missing partition key
        """
        return Partition(self, kwargs=kwargs)


class Partition:
    def __init__(self, partitioning: Partitioning, kwargs: Dict[str, Any]):
        # FIXME: check to make sure that kwargs match partition key names and types
        self.partitioning = partitioning
        self.kwargs = kwargs

    def to_path(self):
        """
        Build a partial file path with the partition information
        """
        return os.path.join(*(str(self.kwargs[name]) for name in self.partitioning.names))


import argparse
import io
import json
import struct
import sys

import flatbuffers
from meillionen.interface.method_request import MethodRequest
from meillionen.interface.module_interface import ModuleInterface


class Server:
    """
    A class that handles method requests from clients

    Currently this is only for handling requests from stdin but
    in the future it may handle requests over http2.
    """

    def __init__(self, module: ModuleInterface):
        self.module = module

    def _interface(self, kwargs):
        builder = flatbuffers.Builder()
        builder.Finish(self.module.serialize(builder))
        interface = builder.Output()
        sys.stdout.buffer.write(interface)

    def _describe(self, kwargs):
        return {
            'list': self._describe_list(),
            'detail': self._describe_detail(kwargs['class_name'])
        }[kwargs['describe_command']]

    def _describe_list(self):
        json.dump(list(self.module.classes.keys()), sys.stdout)

    def _describe_detail(self, class_name: str):
        # Not Implemented Yet
        json.dump(self.module.classes[class_name].metadata, sys.stdout)

    def _run(self, kwargs, fd: io.BytesIO = sys.stdin.buffer):
        size_s = fd.read(4)
        size = struct.unpack('<I', size_s)[0]
        buffer = fd.read(size)
        req = MethodRequest.deserialize(buffer)
        self.run(req)

    def run(self, req: MethodRequest):
        self.module(req)

    def _build_cli(self):
        p = argparse.ArgumentParser()
        sp = p.add_subparsers()

        interface_p = sp.add_parser('interface')
        interface_p.set_defaults(command='interface')

        describe_p = sp.add_parser('describe')
        describe_p.set_defaults(command='describe')
        describe_sp = describe_p.add_subparsers()

        describe_list_p = describe_sp.add_parser('list')
        describe_list_p.set_defaults(describe_command='list')

        describe_detail_p = describe_sp.add_parser('detail')
        describe_detail_p.set_defaults(describe_command='detail')
        describe_detail_p.add_argument('class_name')

        run_p = sp.add_parser('run')
        run_p.set_defaults(command='run')

        return p

    def cli(self):
        """
        Returns an argparse cli to access the model with
        """
        args = self._build_cli().parse_args()
        if hasattr(args, 'command'):
            command = args.command
            {
                'interface': self._interface,
                'describe': self._describe,
                'run': self._run
            }[command](args)



class ClassNotFound(KeyError):
    pass


class MethodNotFound(KeyError):
    pass


class ResourceNotFound(KeyError):
    pass


class HandlerNotFound(KeyError):
    pass


class ExtensionHandlerNotFound(KeyError):
    pass


from typing import Any, List, Dict

import flatbuffers

from . import _FunctionInterface as fi
from .schema import Schema, _Schema
from .base import deserialize_to_dict, serialize_dict, FlatbufferMixin
from meillionen.interface.mutability import Mutability


def default_handler(**kwargs):
    raise NotImplemented()


class _MethodInterface(fi._FunctionInterface, FlatbufferMixin):
    ARGS_OFFSET = 6

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        return cls.get_root_as(buf, offset)

    def Args(self, j):
        return self._get_resource(j, self.ARGS_OFFSET, _Schema)


class MethodInterface:
    """
    A description of a method. It has a name, arguments (with schema information)
    and a handler (which runs the method)
    """
    def __init__(self, name, args: List[Any], handler=default_handler):
        self.name = name
        self.args = {s.name: s for s in args} if not hasattr(args, 'values') else args
        self.handler = handler

    def __call__(self, **kwargs):
        return self.handler(**kwargs)

    @classmethod
    def from_interface(cls, interface: _MethodInterface):
        name = interface.Name().decode('utf-8')
        args = deserialize_to_dict(
            constructor=Schema.from_class,
            getter=interface.Args,
            n=interface.ArgsLength())
        return cls(
            name=name,
            args=args
        )

    @classmethod
    def deserialize(cls, buffer):
        """
        Builds a method interface from a buffer
        """
        interface = _MethodInterface.GetRootAs(buffer, 0)
        return cls.from_interface(interface)

    def serialize(self, builder: flatbuffers.Builder):
        """
        Serializes a method interface into a flatbuffer builder
        """
        name_off = builder.CreateString(self.name)
        args_off = serialize_dict(
            vector_builder=fi.StartArgsVector,
            builder=builder,
            xs=self.args
        )
        fi.Start(builder)
        fi.AddName(builder, name_off)
        fi.AddArgs(builder, args_off)
        return fi.End(builder)

    def process_kwargs(self, kwargs: Dict[str, Any]):
        """
        Preprocesses resource payloads to reduce boilerplate code in the handler
        """
        # FIXME: This method should removed and just pass a request class object to the handler
        result = {}
        for name, resource in kwargs.items():
            handler = self.args[name]
            if handler.mutability == Mutability.read:
                result[name] = handler.load(resource)
            elif handler.mutability == Mutability.write:
                result[name] = (handler, resource)
        return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['meillionen', 'meillionen.interface']

package_data = \
{'': ['*']}

install_requires = \
['flatbuffers>=2.0,<3.0',
 'landlab>=2.3.0,<3.0.0',
 'netCDF4>=1.5.7,<2.0.0',
 'pyarrow>=5.0.0,<6.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'sh>=1.14.2,<2.0.0',
 'xarray>=0.19.0,<0.20.0']

extras_require = \
{'prefect': ['prefect>=0.15.6,<0.16.0']}

setup_kwargs = {
    'name': 'meillionen',
    'version': '0.1.2',
    'description': 'A model interface serialization and rpc framework',
    'long_description': '# Meillionen\n\nMeillionen will be a project to facilitate the interoperability of simulation models written in different languages and frameworks for the physical and social sciences. Current work is focused on combining models with command line interfaces in python but work on bidirectional communication is planned.\n\n## Installation\n\nWith pip\n\n```bash\npip install meillionen\n```\n\n## Development\n\nSetup the environment\n\n```bash\nconda env create -f environment.yml -n meillionen\nconda activate meillionen\n```\n\nRun the tests\n\n```bash\npoetry run pytest\n```\n\nInstall the dependencies\n\n```bash\npoetry install\n```\n',
    'author': 'Calvin Pritchard',
    'author_email': 'calvin.pritchard@asu.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/openmodelingfoundation/meillionen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



class Response:
    """
    Response to a request from the client to call a method
    """
    def __init__(self, handlers, resources):
        self._handlers = handlers
        self._resources = resources

    def handler(self, name: str):
        """
        Retrieve a handler

        :param name: the name of the handler
        """
        return self._handlers[name]

    def resource(self, name: str):
        """
        Retrieve a resource

        :param name: the name of the resource
        """
        return self._resources[name]

    def load(self, name: str):
        """
        Loads a resource using its handler

        :param name: the name of the resource and handler
        """
        handler = self.handler(name)
        resource = self.resource(name)
        return handler.load(resource)

    def save(self, name: str, data):
        """
        Saves a resource using its handler

        :param name: the name of the resource and handler
        :param data: the data to save
        """
        handler = self.handler(name)
        resource = self.resource(name)
        return handler.save(resource, data)


from typing import Dict, List

import flatbuffers

from .base import deserialize_to_dict, serialize_dict, FlatbufferMixin
from .class_interface import ClassInterface, _ClassInterface
from .method_interface import MethodInterface
from .method_request import MethodRequest
from . import _ModuleInterface as mi
from meillionen.exceptions import ClassNotFound


class _ModuleInterface(mi._ModuleInterface, FlatbufferMixin):
    CLASS_OFFSET = 4

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        return cls.get_root_as(buf, offset)

    def Classes(self, j):
        return self._get_resource(j, self.CLASS_OFFSET, _ClassInterface)


class ModuleInterface:
    """
    A collection of class interfaces
    """
    def __init__(self, classes: List[ClassInterface]):
        self.classes = {c.name: c for c in classes} if not hasattr(classes, 'values') else classes

    def __call__(self, req: MethodRequest):
        """
        Call a class method

        :param req: a request to a class method
        """
        try:
            klass = self.classes[req.class_name]
        except KeyError as e:
            raise ClassNotFound() from e

        method = klass.get_method(req.method_name)
        kwargs = method.process_kwargs(req.kwargs)
        return method(**kwargs)

    @classmethod
    def deserialize(cls, buffer):
        """
        Builds a module interface from a buffer
        """
        interface = _ModuleInterface.GetRootAs(buffer, 0)
        classes = deserialize_to_dict(
            constructor=ClassInterface.from_interface,
            getter=interface.Classes,
            n=interface.ClassesLength())
        return cls(classes=classes)

    def serialize(self, builder: flatbuffers.Builder):
        """
        Serialize a module interface into a flatbuffer builder

        :param builder: the flatbuffer builder
        """
        class_off = serialize_dict(
            builder=builder,
            vector_builder=mi.StartClassesVector,
            xs=self.classes
        )
        mi.Start(builder)
        mi.AddClasses(builder, class_off)
        return mi.End(builder)

    def get_method(self, req: MethodRequest) -> MethodInterface:
        klass = self.classes[req.class_name]
        method = klass._methods[req.method_name]
        return method